"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.installDeps = installDeps;
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const config_1 = require("@pnpm/config");
const deps_status_1 = require("@pnpm/deps.status");
const error_1 = require("@pnpm/error");
const get_context_1 = require("@pnpm/get-context");
const filter_workspace_packages_1 = require("@pnpm/filter-workspace-packages");
const manifest_utils_1 = require("@pnpm/manifest-utils");
const workspace_find_packages_1 = require("@pnpm/workspace.find-packages");
const plugin_commands_rebuild_1 = require("@pnpm/plugin-commands-rebuild");
const pnpmfile_1 = require("@pnpm/pnpmfile");
const store_connection_manager_1 = require("@pnpm/store-connection-manager");
const core_1 = require("@pnpm/core");
const logger_1 = require("@pnpm/logger");
const sort_packages_1 = require("@pnpm/sort-packages");
const workspace_pkgs_graph_1 = require("@pnpm/workspace.pkgs-graph");
const workspace_state_1 = require("@pnpm/workspace.state");
const is_subdir_1 = __importDefault(require("is-subdir"));
const errors_1 = require("./errors");
const getPinnedVersion_1 = require("./getPinnedVersion");
const getSaveType_1 = require("./getSaveType");
const nodeExecPath_1 = require("./nodeExecPath");
const recursive_1 = require("./recursive");
const updateWorkspaceDependencies_1 = require("./updateWorkspaceDependencies");
const installConfigDeps_1 = require("./installConfigDeps");
const OVERWRITE_UPDATE_OPTIONS = {
    allowNew: true,
    update: false,
};
async function installDeps(opts, params) {
    if (!opts.update && !opts.dedupe && params.length === 0 && opts.optimisticRepeatInstall) {
        const { upToDate } = await (0, deps_status_1.checkDepsStatus)({
            ...opts,
            ignoreFilteredInstallCache: true,
        });
        if (upToDate) {
            (0, logger_1.globalInfo)('Already up to date');
            return;
        }
    }
    if (opts.workspace) {
        if (opts.latest) {
            throw new error_1.PnpmError('BAD_OPTIONS', 'Cannot use --latest with --workspace simultaneously');
        }
        if (!opts.workspaceDir) {
            throw new error_1.PnpmError('WORKSPACE_OPTION_OUTSIDE_WORKSPACE', '--workspace can only be used inside a workspace');
        }
        if (!opts.linkWorkspacePackages && !opts.saveWorkspaceProtocol) {
            if (opts.rawLocalConfig['save-workspace-protocol'] === false) {
                throw new error_1.PnpmError('BAD_OPTIONS', 'This workspace has link-workspace-packages turned off, \
so dependencies are linked from the workspace only when the workspace protocol is used. \
Either set link-workspace-packages to true or don\'t use the --no-save-workspace-protocol option \
when running add/update with the --workspace option');
            }
            else {
                opts.saveWorkspaceProtocol = true;
            }
        }
        // @ts-expect-error
        opts['preserveWorkspaceProtocol'] = !opts.linkWorkspacePackages;
    }
    let store = await (0, store_connection_manager_1.createOrConnectStoreController)(opts);
    if (opts.configDependencies) {
        await (0, installConfigDeps_1.installConfigDeps)(opts.configDependencies, {
            registries: opts.registries,
            rootDir: opts.lockfileDir ?? opts.rootProjectManifestDir,
            store: store.ctrl,
        });
    }
    if (!opts.ignorePnpmfile && !opts.hooks) {
        opts.hooks = (0, pnpmfile_1.requireHooks)(opts.lockfileDir ?? opts.dir, opts);
        if (opts.hooks.fetchers != null || opts.hooks.importPackage != null) {
            store = await (0, store_connection_manager_1.createOrConnectStoreController)(opts);
        }
    }
    const includeDirect = opts.includeDirect ?? {
        dependencies: true,
        devDependencies: true,
        optionalDependencies: true,
    };
    const forceHoistPattern = typeof opts.rawLocalConfig['hoist-pattern'] !== 'undefined' ||
        typeof opts.rawLocalConfig['hoist'] !== 'undefined';
    const forcePublicHoistPattern = typeof opts.rawLocalConfig['shamefully-hoist'] !== 'undefined' ||
        typeof opts.rawLocalConfig['public-hoist-pattern'] !== 'undefined';
    const allProjects = opts.allProjects ?? (opts.workspaceDir
        ? await (0, workspace_find_packages_1.findWorkspacePackages)(opts.workspaceDir, { ...opts, patterns: opts.workspacePackagePatterns })
        : []);
    if (opts.workspaceDir) {
        const selectedProjectsGraph = opts.selectedProjectsGraph ?? selectProjectByDir(allProjects, opts.dir);
        if (selectedProjectsGraph != null) {
            const sequencedGraph = (0, sort_packages_1.sequenceGraph)(selectedProjectsGraph);
            // Check and warn if there are cyclic dependencies
            if (!opts.ignoreWorkspaceCycles && !sequencedGraph.safe) {
                const cyclicDependenciesInfo = sequencedGraph.cycles.length > 0
                    ? `: ${sequencedGraph.cycles.map(deps => deps.join(', ')).join('; ')}`
                    : '';
                if (opts.disallowWorkspaceCycles) {
                    throw new error_1.PnpmError('DISALLOW_WORKSPACE_CYCLES', `There are cyclic workspace dependencies${cyclicDependenciesInfo}`);
                }
                logger_1.logger.warn({
                    message: `There are cyclic workspace dependencies${cyclicDependenciesInfo}`,
                    prefix: opts.workspaceDir,
                });
            }
            let allProjectsGraph;
            if (opts.dedupePeerDependents) {
                allProjectsGraph = opts.allProjectsGraph ?? (0, workspace_pkgs_graph_1.createPkgGraph)(allProjects, {
                    linkWorkspacePackages: Boolean(opts.linkWorkspacePackages),
                }).graph;
            }
            else {
                allProjectsGraph = selectedProjectsGraph;
                if (!allProjectsGraph[opts.workspaceDir]) {
                    allProjectsGraph = {
                        ...allProjectsGraph,
                        ...selectProjectByDir(allProjects, opts.workspaceDir),
                    };
                }
            }
            await recursiveInstallThenUpdateWorkspaceState(allProjects, params, {
                ...opts,
                ...(0, config_1.getOptionsFromRootManifest)(opts.rootProjectManifestDir, opts.rootProjectManifest ?? {}),
                forceHoistPattern,
                forcePublicHoistPattern,
                allProjectsGraph,
                selectedProjectsGraph,
                storeControllerAndDir: store,
                workspaceDir: opts.workspaceDir,
            }, opts.update ? 'update' : (params.length === 0 ? 'install' : 'add'));
            return;
        }
    }
    // `pnpm install ""` is going to be just `pnpm install`
    params = params.filter(Boolean);
    const dir = opts.dir || process.cwd();
    let workspacePackages;
    if (opts.workspaceDir) {
        workspacePackages = (0, get_context_1.arrayOfWorkspacePackagesToMap)(allProjects);
    }
    let { manifest, writeProjectManifest } = await (0, cli_utils_1.tryReadProjectManifest)(opts.dir, opts);
    if (manifest === null) {
        if (opts.update === true || params.length === 0) {
            throw new error_1.PnpmError('NO_PKG_MANIFEST', `No package.json found in ${opts.dir}`);
        }
        manifest = {};
    }
    const installOpts = {
        ...opts,
        ...(0, config_1.getOptionsFromRootManifest)(opts.dir, (opts.dir === opts.rootProjectManifestDir ? opts.rootProjectManifest ?? manifest : manifest)),
        forceHoistPattern,
        forcePublicHoistPattern,
        // In case installation is done in a multi-package repository
        // The dependencies should be built first,
        // so ignoring scripts for now
        ignoreScripts: !!workspacePackages || opts.ignoreScripts,
        linkWorkspacePackagesDepth: opts.linkWorkspacePackages === 'deep' ? Infinity : opts.linkWorkspacePackages ? 0 : -1,
        sideEffectsCacheRead: opts.sideEffectsCache ?? opts.sideEffectsCacheReadonly,
        sideEffectsCacheWrite: opts.sideEffectsCache,
        storeController: store.ctrl,
        storeDir: store.dir,
        workspacePackages,
    };
    if (opts.global && opts.pnpmHomeDir != null) {
        const nodeExecPath = await (0, nodeExecPath_1.getNodeExecPath)();
        if ((0, is_subdir_1.default)(opts.pnpmHomeDir, nodeExecPath)) {
            installOpts['nodeExecPath'] = nodeExecPath;
        }
    }
    let updateMatch;
    if (opts.update) {
        if (params.length === 0) {
            const ignoreDeps = manifest.pnpm?.updateConfig?.ignoreDependencies;
            if (ignoreDeps?.length) {
                params = (0, recursive_1.makeIgnorePatterns)(ignoreDeps);
            }
        }
        updateMatch = params.length ? (0, recursive_1.createMatcher)(params) : null;
    }
    else {
        updateMatch = null;
    }
    if (updateMatch != null) {
        params = (0, recursive_1.matchDependencies)(updateMatch, manifest, includeDirect);
        if (params.length === 0) {
            if (opts.latest)
                return;
            if (opts.depth === 0) {
                throw new error_1.PnpmError('NO_PACKAGE_IN_DEPENDENCIES', 'None of the specified packages were found in the dependencies.');
            }
        }
    }
    if (opts.update && opts.latest && (!params || (params.length === 0))) {
        params = Object.keys((0, manifest_utils_1.filterDependenciesByType)(manifest, includeDirect));
    }
    if (opts.workspace) {
        if (!params || (params.length === 0)) {
            params = (0, updateWorkspaceDependencies_1.updateToWorkspacePackagesFromManifest)(manifest, includeDirect, workspacePackages);
        }
        else {
            params = (0, updateWorkspaceDependencies_1.createWorkspaceSpecs)(params, workspacePackages);
        }
    }
    if (params?.length) {
        const mutatedProject = {
            allowNew: opts.allowNew,
            binsDir: opts.bin,
            dependencySelectors: params,
            manifest,
            mutation: 'installSome',
            peer: opts.savePeer,
            pinnedVersion: (0, getPinnedVersion_1.getPinnedVersion)(opts),
            rootDir: opts.dir,
            targetDependenciesField: (0, getSaveType_1.getSaveType)(opts),
        };
        const { updatedProject, ignoredBuilds } = await (0, core_1.mutateModulesInSingleProject)(mutatedProject, installOpts);
        if (opts.save !== false) {
            await writeProjectManifest(updatedProject.manifest);
        }
        if (!opts.lockfileOnly) {
            await (0, workspace_state_1.updateWorkspaceState)({
                allProjects,
                settings: opts,
                workspaceDir: opts.workspaceDir ?? opts.lockfileDir ?? opts.dir,
                pnpmfileExists: opts.hooks?.calculatePnpmfileChecksum != null,
                filteredInstall: allProjects.length !== Object.keys(opts.selectedProjectsGraph ?? {}).length,
                configDependencies: opts.configDependencies,
            });
        }
        if (opts.strictDepBuilds && ignoredBuilds?.length) {
            throw new errors_1.IgnoredBuildsError(ignoredBuilds);
        }
        return;
    }
    const { updatedManifest, ignoredBuilds } = await (0, core_1.install)(manifest, installOpts);
    if (opts.update === true && opts.save !== false) {
        await writeProjectManifest(updatedManifest);
    }
    if (opts.strictDepBuilds && ignoredBuilds?.length) {
        throw new errors_1.IgnoredBuildsError(ignoredBuilds);
    }
    if (opts.linkWorkspacePackages && opts.workspaceDir) {
        const { selectedProjectsGraph } = await (0, filter_workspace_packages_1.filterPkgsBySelectorObjects)(allProjects, [
            {
                excludeSelf: true,
                includeDependencies: true,
                parentDir: dir,
            },
        ], {
            workspaceDir: opts.workspaceDir,
        });
        await recursiveInstallThenUpdateWorkspaceState(allProjects, [], {
            ...opts,
            ...OVERWRITE_UPDATE_OPTIONS,
            allProjectsGraph: opts.allProjectsGraph,
            selectedProjectsGraph,
            workspaceDir: opts.workspaceDir, // Otherwise TypeScript doesn't understand that is not undefined
        }, 'install');
        if (opts.ignoreScripts)
            return;
        await (0, plugin_commands_rebuild_1.rebuildProjects)([
            {
                buildIndex: 0,
                manifest: await (0, cli_utils_1.readProjectManifestOnly)(opts.dir, opts),
                rootDir: opts.dir,
            },
        ], {
            ...opts,
            pending: true,
            storeController: store.ctrl,
            storeDir: store.dir,
            skipIfHasSideEffectsCache: true,
        });
    }
    else {
        if (!opts.lockfileOnly) {
            await (0, workspace_state_1.updateWorkspaceState)({
                allProjects,
                settings: opts,
                workspaceDir: opts.workspaceDir ?? opts.lockfileDir ?? opts.dir,
                pnpmfileExists: opts.hooks?.calculatePnpmfileChecksum != null,
                filteredInstall: allProjects.length !== Object.keys(opts.selectedProjectsGraph ?? {}).length,
                configDependencies: opts.configDependencies,
            });
        }
    }
}
function selectProjectByDir(projects, searchedDir) {
    const project = projects.find(({ rootDir }) => path_1.default.relative(rootDir, searchedDir) === '');
    if (project == null)
        return undefined;
    return { [searchedDir]: { dependencies: [], package: project } };
}
async function recursiveInstallThenUpdateWorkspaceState(allProjects, params, opts, cmdFullName) {
    const recursiveResult = await (0, recursive_1.recursive)(allProjects, params, opts, cmdFullName);
    if (!opts.lockfileOnly) {
        await (0, workspace_state_1.updateWorkspaceState)({
            allProjects,
            settings: opts,
            workspaceDir: opts.workspaceDir,
            pnpmfileExists: opts.hooks?.calculatePnpmfileChecksum != null,
            filteredInstall: allProjects.length !== Object.keys(opts.selectedProjectsGraph ?? {}).length,
            configDependencies: opts.configDependencies,
        });
    }
    return recursiveResult;
}
//# sourceMappingURL=installDeps.js.map